	The Fragmentation Inhibiter

Normally, VMS disks become fragmented quickly during use. This 
produces many extra index file accesses and disk reads, slowing down
file access.

Oddly enough, this fragmentation occurs because of a VMS optimization.
When you delete a file, VMS does not instantly zero the bitmap bits
corresponding to its space, but leaves it in an "extent cache" so that
allocating the space to a new file can mean VMS need not visit some
bitmap areas at all. This can in some cases speed disk access.

The problem with the strategy is that almost all files grow by
allocating several clusters and returning the extras by truncation
after closing. This means the extent cache is generally full of
tiny fragments of disk storage, so that new files are then allocated
out of this "chaff" and wind up extremely fragmented from the start
if default allocation is allowed. This can be defeated if we tell VMS
to allocate files using big chunks of space first, so the chaff can
be reclaimed later, or at worst concentrated in a few files.

It also often happens that some files (e.g. log files) can stay open a
long time and grow by grabbing space a few blocks at a time. When a file
is 10,000 blocks long, though, it is generally not wise to extend it by
2 blocks at a time. Rather, it makes sense to allocate some fraction
of the file size (subject to not trying to take too large a piece of
the disk's remaining space at once) for growth. This keeps the file
from having to be extended as often and cuts fragmentation too. Since the
files remain open, this can be the simplest way to reduce disk fragmentation
where many defraggers are unable to defragment open files.

The Fragmentation Inhibiter handles both these operations automatically.

The Fragmentation Inhibiter can command file extension to be done
"Contiguous Best Try" when this is possible. This causes VMS to
flush (i.e., forget about) its extent cache and attempt to find
a space big enough to hold the whole requested area before using
badly fragmented store. The program can do this for every extension
(the default and recommended behavior) or every Nth extension,
if you feel the need to have the extent cache used, but cleared
periodically. (There are a few situations, like those on NEWS disks,
where thousands of tiny files are created frequently and in which the
normal extent cache operation can be helpful.)

Fragmentation Inhibiter also controls the amount of space requested
when a program extends a file (i.e., makes it longer). It can
set a minimum extent request, so even explicit requests for one
block at a time can be increased, and is able to request that a
file be extended by 1/N of its current size, subject to free space
constraints on the volume and a maximum extend amount. It will
always request at least as much space as the program wanted, but
will attempt to extend by 1/N of the file size, 1/8 of the available
free space on the disk, or the maximum extent request, whichever
is least.

Suppose, for example, you have a file SYSAUDIT.LOG that is now
1000 blocks long, and suppose you set Fragmentation Inhibiter to
extend by 1/4 of the file size. Suppose also your volume default
allocation is 10 blocks.

Now if the file is extended, Fragmentation Inhibiter will cause it to
grow to 1000 + 250 + 1 blocks long in one extend operation. This
will be usually found contiguously and thus the file gains one
extent. 

If normal VMS allocation had been in effect, the file would have been
extended about 25 times to gain this much space, would probably be
multi-header if it wasn't already, and could not be defragmented by
conventional techniques because it would still be open.

As a file grows, the number of extend operations needed for it
to grow to a given size drops quickly. Consider the following table,
which assumes volume default allocation of 20 blocks and F.I. set
to extend by 1/4 of file size:

   Extension		Size with F.I.		Size without F.I.

	1			20			20
	2			40			40
	3			60			60
	4			80			80
	5		       101		       100
	6		       127		       120
	7		       159		       140
	8		       199		       160
	9		       249		       180
       10		       312		       200
       11		       391		       220
       12		       489                     240
       13		       612		       260
       14		       766		       280
       15		       958		       300
       16		      1198		       320

This clearly shows that less extensions can handle a given file
size in this way. If F.I. were not there, the file would need
to be extended another 44 times to hold 1198 blocks.

File extension is an expensive operation; by arranging it to be less
often needed, F.I. can permit file writing to enjoy up to a 30%
speed gain, in addition to permitting faster access to files written with
it. This is the gain from not having to do the extend operation
as often.

SUMMARY:

Fragmentation Inhibiter permits disk fragmentation to be controlled
by altering somewhat the VMS file extend operation in a way
that greatly slows the refragmentation of disk space. The product
can be installed on any disks where this is desired and its
parameters can be set on a per-disk basis. Disks where the 
Fragmentation Inhibiter is not installed are unaffected.

Installation:

From SYSTEM (or some other fully privileged account) type

$@SYS$UPDATE:VMSINSTAL SAFETY010

The system will install Fragmentation Inhibiter for you. Then use
the FI_SETUP.COM script to generate a script SYS$MANAGER:FI_STARTUP.COM
which can be run from your systartup_v5.com or systartup_vms.com
at boot.

Use:

First, issue the command

$SET COMMAND SYS$SYSTEM:JTDMN

Then for every disk whose fragmentation is to be controlled, you
must connect a unit of the JT: pseudodevice. The SYSGEN connect command
looks like this:

$ sysgen connect JTA3:/noada/driver=jtdriver

Then you use JTDMN to point another disk at it for Frag Inhibiting:

e.g.

$ SET COMM SYS$SYSTEM:JTDMN

$ JTDMN/FCNMSK:2/MODIFY JTA3: DKA500:


Now DKA500: will have its fragmentation inhibited.

There are a number of controls available. These use the gene4ral command
format:

$JTDMN/switches JTAn: FragInhibited_device:

where

   JTAn: is a unit of JTdriver, one unit per victim disk.
   FragInhibited_device: is the devicename of a disk having its fragmentation
	inhibited (by preventing it in the first place).

Switches are as follows:

/FRACTION:N	Sets fraction of file size to extend by. If n=3
		(JTDMN/FRACTION:3 JTAn: DEVICE:) then file
		extension will be by 1/3 of the file size, subject
		to other constraints. Default is N=4 to extend by
		1/4 of file size.
/CBT:N		Do file extension contiguous best try every Nth
		open. Default is 1 so file extension is always tried
		contiguous best try. You can set the Frag Inhibiter
		to do this only every Nth time (N up to 1,000,000,000)
		to allow use of the extent cache some if you like.
		Note that contiguous extensions are left alone.
/MINIMUM:N	Sets minimum extension request; default is 10
		blocks. This is sometimes useful for programs
		which explicitly request very small extents.
/MAXIMUM:N	Sets maximum extension request. If 1/fraction of file
		size is greater than N, then we use this maximum to
		extend the file. Thus, setting /MAX:10000 would
		mean we never request more than 10000 blocks
		extension even on huge files (though if the 
		program requests more, we NEVER reduce an extension
		request, ever.)
/ALDEFONLY	If specified, modifies extensions ONLY where
		the program says to use volume default extension.
		Otherwise always modifies extensions provided
		the modified extension is at least as large
		as that originally requested.
			Using this setting is extremely "safe" in that
		programs requesting default extension are "expecting"
		the extension not to be exactly known, but many 
		"problem" programs have hardcoded 2 or 3 block extension
		lengths and will continue to cause fragmentation and
		use lots of extend requests if this switch is set.
		Therefore it is a per-disk site selection which to 
		use.

Ensure that the /FCNMSK:N switch is present with a value N that contains the
		2 bit when using JTdriver to Inhibit fragmentation. Also be
		sure to add the switch /MODIFY to inhibit JTDMN looping if
		only using frag avoidance. The /MODIFY switch just tells
		the program to exit once its modifications are done. It
		is used where all that is happening is fragmentation
		control, since that requires no lasting presence of a
		daemon.

	In all cases, the Fragmentation Inhibiter will never extend a
	file by more than 1/8 of the free space on the volume and
	will never cause an extend request to have less blocks requested
	than the program asked for. It will not reduce the program's
	request even if the freespace constraint is violated, nor will
	it touch contiguous extend requests.


	The following switch can be used if needed. It MUST be used
	with CAUTION since the correct JF unit must be given.

/DEASSIGN 	Remove the JTAn: unit association with the target
		device and turn off frag Inhibit. Format of
		this command is:

	$ JTDMN/DEASSIGN JTAn: FragInhibited_device:

	Note that the unit of JTAn: that is specified MUST be the same as
	that which was used to initially start fragmentation inhibiting.
	If the wrong unit is used, the system will refuse to deassign
	the unit (and leave the JF unit online). This safeguard will
	prevent system damage, but the Fragmentation Inhibiter should be
	set the way you prefer and left alone for normal use.


(Note: This documentation is preliminary and may be expected to be revised
somewhat. In particular, JTdmn will become some other name eventually
and the current JTdmn has other functions too, which would be removed
from the F.I. version so that it will exit after each setup command.)
